import logging

from redbaron import RedBaron


def fix_access_controls(red: RedBaron):
    # Check for DAG access controls
    dag_access_control_nodes = red.find_all("NameNode", value="access_control")
    if len(dag_access_control_nodes):
        accts = []
        for node in dag_access_control_nodes:
            accts.append(node.parent.dict_.dictitem_.key_.dumps())
            arg = node.parent
            arg.value = "None"
        logging.warning(f"Contains DAG access controls. Setting to None: {accts}")



from redbaron import Node, RedBaron


def fix_xcom_push(red: RedBaron):
    # Fix xcom_push -> do_xcom_push in Operator args
    xcom_push_nodes = red.find_all("NameNode", value="xcom_push")
    args_xcom_push_nodes = xcom_push_nodes.filter(
        lambda x: x.parent.type == "call_argument"
    )
    args_xcom_push_nodes.map(lambda x: x.replace("do_xcom_push"))


def fix_xcom_pull(red: RedBaron):
    # Fix xcom_push -> do_xcom_push in Operator args
    xcom_pull_nodes = red.find_all("NameNode", value="xcom_pull")
    args_xcom_pull_nodes = xcom_pull_nodes.filter(
        lambda x: x.parent.type == "call_argument"
    )
    for node in args_xcom_pull_nodes:
        arg = node.parent
        _remove_arg_via_fst(arg)


# Method to resolve https://github.com/PyCQA/redbaron/issues/100
def _remove_arg_via_fst(arg):
    call_fst = arg.parent.fst()
    # Account for commas in Call FST
    # Last argument or last argument with trailing comma
    if (
        len(call_fst["value"]) == arg.index_on_parent * 2
        or len(call_fst["value"]) == arg.index_on_parent * 2 + 1
    ):
        call_fst["value"] = call_fst["value"][: arg.index_on_parent * 2]
    # Not last argument
    else:
        call_fst["value"] = (
            call_fst["value"][: arg.index_on_parent * 2]
            + call_fst["value"][arg.index_on_parent * 2 + 2 :]
        )
    new_call_node = Node.from_fst(
        call_fst, parent=arg.parent.parent, on_attribute="call"
    )
    new_call_node.parent.call_.replace(new_call_node)



import importlib
import inspect
import logging
import os
import pkgutil
import sys
import warnings
from typing import List, Optional

from flake8.api import legacy as flake8

from redbaron import RedBaron
from baron.parser import ParsingError

from fixers import fix_access_controls, fix_imports, fix_xcom_pull, fix_xcom_push

# logging.basicConfig(level=logging.DEBUG)

warnings.filterwarnings("error")

LOG_FORMAT = "[%(name)s] [%(levelname)s] %(message)s"
LOG_LEVEL = logging.INFO
formatter = logging.Formatter(LOG_FORMAT)
nova_logger = logging.getLogger("nova")
nova_logger.setLevel(LOG_LEVEL)
handler = logging.StreamHandler()
handler.setLevel(LOG_LEVEL)
handler.setFormatter(formatter)
nova_logger.addHandler(handler)

# TODO Give control over fixers to run
_fixers = [
    "fix_imports",
    "fix_access_controls",
    "fix_xcom_push",
    "fix_xcom_pull",
    "update_dag_name",
    # TODO Worker Queues
]
_reports = [
    # TODO Async Optimizations
    # TODO Worker Queues
]

common_dag_class_imports = [
    "airflow.hooks",
    "airflow.operators",
    "airflow.sensors",
    "airflow.providers",
]


def _get_class_import_map(root_module) -> dict:
    found_packages = list(pkgutil.iter_modules(root_module.__path__))
    if not found_packages:
        return {}
    # action_names = [
    #     (name.split(".")[-1], ispkg) for _, name, ispkg in found_packages
    # ]
    # importer = found_packages[0][0]
    class_import_map = {}
    for x, name, ispkg in found_packages:
        if ispkg:
            try:
                module = importlib.import_module(f"{root_module.__name__}.{name}")
            except Exception as e:
                breakpoint()
            submodule_class_import_map = _get_class_import_map(module)
            class_import_map = {**submodule_class_import_map, **class_import_map}

        else:
            try:
                module = importlib.import_module(f"{root_module.__name__}.{name}")
                if f"{root_module.__name__}.{name}" not in sys.modules:
                    continue
                classes = inspect.getmembers(module, inspect.isclass)
                for class_name, c in classes:
                    if c.__module__ != f"{root_module.__name__}.{name}":
                        continue
                    class_import_map[class_name] = {
                        "value": f"{root_module.__name__}.{name}",
                        "targets": class_name,
                    }
            except DeprecationWarning as err:
                nova_logger.debug(
                    f"Skip deprecated module {root_module.__name__}.{name}: {err}"
                )
            except ModuleNotFoundError as err:
                nova_logger.error(
                    f"Cannot load modules from {root_module.__name__}.{name}: {err}"
                )
            except UserWarning as err:
                continue
    return class_import_map


def _generate_import_map() -> dict:
    class_import_map = {}
    for common_dag_class_import in common_dag_class_imports:
        root_module = importlib.import_module(common_dag_class_import)
        class_import_map = {
            **_get_class_import_map(root_module),
            **class_import_map,
        }
    return class_import_map


def _upgrade_dag_file(
    filename: str, out_dir: str, class_import_map, rename_dag: Optional[bool] = False
) -> None:
    if "upgraded" in filename:
        return
    red = _load_dag_file(filename)
    if red is None:
        return
    fix_imports(class_import_map, red)
    fix_access_controls(red)
    fix_xcom_push(red)
    fix_xcom_pull(red)
    if rename_dag:
        _update_dag_name(red)
    updated_filename = _write_updated_dag_file(filename, out_dir, red)
    _validate_updated_dag(updated_filename)


def _load_dag_file(filename: str) -> RedBaron:
    try:
        with open(filename, "r") as f:
            code = f.read()
        return RedBaron(code)
    except ParsingError as pe:
        nova_logger.error(f'{filename} contains errors and is not valid Python.\n {pe}')


def _write_updated_dag_file(filename: str, out_dir: str, red: RedBaron) -> str:
    upgraded_filename = filename
    upgraded_filename = os.path.join(out_dir, os.path.basename(upgraded_filename))

    with open(upgraded_filename, "w") as f:
        f.write(red.dumps())

    return upgraded_filename


def _validate_updated_dag(filename: str) -> None:
    style_guide = flake8.get_style_guide(ignore=[])
    report = style_guide.check_files([filename])
    if len(report.get_statistics('E999')) > 0:
        nova_logger.error(f'{filename} contains errors and is not valid Python.')


def _update_dag_name(red: RedBaron) -> None:
    # update dag name
    dag_def = red.find("NameNode", value="DAG").parent
    dag_name = dag_def.call_[0].string_.value
    file_name = dag_name[:-1]
    ext = dag_name[-1]
    dag_def.call_[0].value = f"{file_name}_upgraded{ext}"


def _create_out_dir_if_not_exists(out_dir: str) -> None:
    os.makedirs(out_dir, exist_ok=True)


def _get_dag_files_from_dir(dag_dir: str) -> List[str]:
    return [
        os.path.join(dag_dir, f)
        for f in os.listdir(dag_dir)
        if os.path.isfile(os.path.join(dag_dir, f)) and f[-3:] == ".py"
    ]


def upgrade_dag_files(
    dag_dir: str,
    out_dir: Optional[str] = None,
    rename_dag: Optional[bool] = False,
):
    if not dag_dir:
        nova_logger.error("No DAG directory provided.")
        exit()
    if not out_dir:
        out_dir = os.path.join(dag_dir, "upgraded/")

    _create_out_dir_if_not_exists(out_dir)
    class_import_map = _generate_import_map()

    dag_files = _get_dag_files_from_dir(dag_dir)
    for dag_file in dag_files:
        _upgrade_dag_file(dag_file, out_dir, class_import_map, rename_dag)


def upgrade_dag_file(
    dag_file: str,
    out_dir: Optional[str] = None,
    rename_dag: Optional[bool] = False,
):
    _create_out_dir_if_not_exists(out_dir)
    class_import_map = _generate_import_map()
    _upgrade_dag_file(dag_file, out_dir, class_import_map, rename_dag)



import logging
import os.path

import click

from nova import upgrade_dag_file, upgrade_dag_files


@click.command()
@click.argument("dag_path", default="dags/")
@click.argument("out_dir", default="dags/upgraded")
@click.option(
    "-r",
    "--rename-dag",
    "rename_dag",
    is_flag=True,
    show_default=True,
    default=False,
    help="Append `_updated` to dag_id.",
)
def cli(dag_path: str, out_dir: str, rename_dag):
    """
    Upgrade all dags in `dag_path` and saves them to `out_dir`.
    :param ctx: Click context
    :param dag_path: "The path to the `dags` folder."
    :param out_dir: "The output directory to store upgraded `dags`."
    :param rename_dag: "Append `_updated` to dag_id."
    :return: None
    """

    LOG_FORMAT = "[%(name)s] [%(levelname)s] %(message)s"
    LOG_LEVEL = logging.INFO
    formatter = logging.Formatter(LOG_FORMAT)
    airflow_logger = logging.getLogger("airflow_modules")
    airflow_logger.setLevel(LOG_LEVEL)
    handler = logging.StreamHandler()
    handler.setLevel(LOG_LEVEL)
    handler.setFormatter(formatter)
    airflow_logger.addHandler(handler)

    if os.path.isfile(dag_path):
        upgrade_dag_file(dag_path, out_dir, rename_dag)
    else:
        upgrade_dag_files(dag_path, out_dir, rename_dag)



import logging

from redbaron import RedBaron


def fix_imports(class_import_map, red: RedBaron):
    import_nodes = red.find_all("FromImportNode")
    import_nodes = import_nodes.filter(lambda x: x.name.dumps() == "airflow")
    # it might be better only act against certain import paths like
    #   contrib_imports = [node for node in import_nodes if node.find("name", value="contrib")]
    # but for now acting on all imports
    for import_node in import_nodes:

        if len(import_node.targets) == 1:
            imported_class = import_node.targets[0].value
            new_imported_class = class_import_map.get(imported_class, None)
            if not new_imported_class:
                logging.warning(f"Unhandled import: {import_node.dumps()}")
                continue

            import_node.value = new_imported_class["value"]
        else:
            had_import_error = False
            for node in import_node.targets:

                imported_class = node.value
                new_imported_class = class_import_map.get(imported_class, None)
                if not new_imported_class:
                    had_import_error = True
                    logging.warning(f"Unhandled import: {import_node.dumps()}")
                    continue
                import_node.insert_after(
                    f"from {new_imported_class['value']} import {imported_class}"
                )
            if not had_import_error:
                # RedBaron doesn't handle weird indentation well
                # It's possible this throws the Exception
                #   Exception: It appears that you have indentation in your CommaList,
                #   for now RedBaron doesn't know how to handle this situation
                #   (which requires a lot of work), sorry about that. You can find
                #   more information here https://github.com/PyCQA/redbaron/issues/100
                del import_node.parent[import_node.index_on_parent]
